@@ -4,7 +4,6 @@ from __future__ import annotations
 # that defines this class.
 # https://stackoverflow.com/a/49872353
 # Will no longer be needed in Python 3.10.
-
 import os
 import random
 import time
@@ -12,23 +11,21 @@ from contextlib import contextmanager
 from datetime import datetime
 from io import UnsupportedOperation
 
-from azure.storage.blob import ContainerClient, BlobClient, BlobLeaseClient
-
 # from azure.storage.blob.aio import (
 # ContainerClient as aContainerClient,
 # BlobClient as aBlobClient,
 # BlobLeaseClient as aBlobLeaseClient,
 # )
 from azure.core.exceptions import (
-    ResourceNotFoundError,
-    ResourceExistsError,
     HttpResponseError,
+    ResourceExistsError,
+    ResourceNotFoundError,
 )
+from azure.storage.blob import BlobClient, BlobLeaseClient, ContainerClient
 from overrides import overrides
 
-from ._upath import LockAcquireError, FileInfo, Upath
-from ._blob import BlobUpath
-from ._local import LocalUpath
+from ._blob import BlobUpath, LocalPathType, _resolve_local_path
+from ._upath import FileInfo, LockAcquireError, Upath
 
 # End user may want to do this:
 # logging.getLogger("azure.storage").setLevel(logging.WARNING)
@@ -55,7 +52,6 @@ class AzureBlobUpath(BlobUpath):
         self,
         *paths: str,
         container_name: str = None,
-        **kwargs,
     ):
         if container_name is None:
             assert len(paths) == 1
@@ -71,7 +67,7 @@ class AzureBlobUpath(BlobUpath):
                 container_name = path[:k]
                 paths = (path[k:],)
 
-        super().__init__(*paths, **kwargs)
+        super().__init__(*paths)
 
         self._container_name = container_name
 
@@ -101,70 +97,19 @@ class AzureBlobUpath(BlobUpath):
     def __str__(self) -> str:
         return f"{self._container_name}://{self._path.lstrip('/')}"
 
-    def __eq__(self, other) -> bool:
-        if other.__class__ is not self.__class__:
-            return NotImplemented
-        if other._container_name != self._container_name:
-            return NotImplemented
-        return self._path == other._path
-
-    def __lt__(self, other) -> bool:
-        if other.__class__ is not self.__class__:
-            return NotImplemented
-        if other._container_name != self._container_name:
-            return NotImplemented
-        return self._path < other._path
-
-    def __le__(self, other) -> bool:
-        if other.__class__ is not self.__class__:
-            return NotImplemented
-        if other._container_name != self._container_name:
-            return NotImplemented
-        return self._path <= other._path
-
-    def __gt__(self, other) -> bool:
-        if other.__class__ is not self.__class__:
-            return NotImplemented
-        if other._container_name != self._container_name:
-            return NotImplemented
-        return self._path > other._path
-
-    def __ge__(self, other) -> bool:
-        if other.__class__ is not self.__class__:
-            return NotImplemented
-        if other._container_name != self._container_name:
-            return NotImplemented
-        return self._path >= other._path
+    @overrides
+    def as_uri(self) -> str:
+        # TODO: what's the conventional lead word for Azure?
+        return f"azure://{self._container_name}/{self._path.lstrip('/')}"
 
     @property
     def container_name(self):
         return self._container_name
 
-    def _copy_file_from(self, source, *, overwrite=False):
-        # TODO: use `overwrite`
-        with self._provide_blob_client():
-            with source._provide_blob_client():
-                copy = self._blob_client.start_copy_from_url(
-                    source._blob_client.url,
-                    requires_sync=True,
-                )
-                assert copy["copy_status"] == "success"
-
     @overrides
-    def _copy_file(self, target: AzureBlobUpath, *, overwrite=False):
-        target._copy_file_from(self, overwrite=overwrite)
-
-    @overrides
-    def export_file(self, target: Upath, *, overwrite=False) -> None:
-        if not isinstance(target, LocalUpath):
-            return super().export_file(target, overwrite=overwrite)
+    def is_file(self) -> bool:
         with self._provide_blob_client():
-            # TODO: check behavior of `download_blob` about
-            # overwrite.
-            os.makedirs(str(target.parent), exist_ok=True)
-            with open(str(target), "wb") as f:
-                data = self._blob_client.download_blob()
-                data.readinto(f)
+            return self._blob_client.exists()
 
     @overrides
     def file_info(self):
@@ -186,23 +131,64 @@ class AzureBlobUpath(BlobUpath):
         except ResourceNotFoundError:
             return None
 
+    @property
     @overrides
-    def import_file(self, source: Upath, *, overwrite: bool = False):
-        if not isinstance(source, LocalUpath):
-            return super().import_file(source, overwrite=overwrite)
-        if not overwrite and self.is_file():
-            # TODO: check the behavior of `upload_blob` related to
-            # behavior about overwrite.
-            raise FileExistsError(self)
+    def root(self) -> AzureBlobUpath:
+        """
+        Return a new path representing the root of the same container.
+        """
+        return self.__class__(
+            container_name=self._container_name,
+        )
+
+    def _copy_file_from(self, source, *, overwrite=False):
+        # TODO: use `overwrite`
+        with self._provide_blob_client():
+            with source._provide_blob_client():
+                copy = self._blob_client.start_copy_from_url(
+                    source._blob_client.url,
+                    requires_sync=True,
+                )
+                assert copy["copy_status"] == "success"
+
+    @overrides
+    def _copy_file(self, target: Upath, *, overwrite=False):
+        if isinstance(target, AzureBlobUpath):
+            target._copy_file_from(self, overwrite=overwrite)
+        else:
+            super()._copy_file(target, overwrite=overwrite)
+
+    @overrides
+    def download_file(self, target: LocalPathType, *, overwrite=False) -> None:
+        target = _resolve_local_path(target)
+        if target.is_file():
+            if not overwrite:
+                raise FileExistsError(str(target))
+            target.remove_file()
+        elif target.is_dir():
+            raise IsADirectoryError(str(target))
+
+        with self._provide_blob_client():
+            # TODO: check behavior of `download_blob` about
+            # overwrite.
+            os.makedirs(str(target.parent), exist_ok=True)
+            with open(str(target), "wb") as f:
+                data = self._blob_client.download_blob()
+                data.readinto(f)
+
+    @overrides
+    def upload_file(self, source: LocalPathType, *, overwrite: bool = False):
+        source = _resolve_local_path(source)
+        if self.is_file():
+            if not overwrite:
+                # TODO: check the behavior of `upload_blob` related to
+                # behavior about overwrite.
+                raise FileExistsError(self)
+            self.remove_file()
         with self._provide_blob_client():
             with open(str(source), "rb") as data:
                 self._blob_client.upload_blob(data)
 
-    @overrides
-    def is_file(self) -> bool:
-        with self._provide_blob_client():
-            return self._blob_client.exists()
-
     @overrides
     def iterdir(self):
         with self._provide_container_client():
@@ -213,7 +199,7 @@ class AzureBlobUpath(BlobUpath):
 
     def _acquire_lease(self, timeout: int = None):
         if timeout is None:
-            timeout = 3600
+            timeout = 300
         t0 = time.perf_counter()
         while True:
             try:
@@ -356,14 +342,6 @@ class AzureBlobUpath(BlobUpath):
             for p in self._container_client.list_blobs(name_starts_with=prefix):
                 yield self / p.name[k:]
 
-    @overrides
-    def with_path(self, *paths):
-        return self.__class__(
-            *paths,
-            container_name=self._container_name,
-            thread_pool_executors=self._thread_pools,
-        )
-
     @overrides
     def write_bytes(self, data: bytes, *, overwrite=False) -> None:
         if self._path == "/":
